--[[

WearMe
	Library for equipping, bagging and moving equipment using a dynamic cache.

By Karl Isenberg (AnduinLothar)

Change Log:

v1.0
- Initial Release

	$Id: WearMe.lua 3705 2006-06-26 08:15:29Z karlkfi $
	$Rev: 3705 $
	$LastChangedBy: karlkfi $
	$Date: 2006-06-26 03:15:29 -0500 (Mon, 26 Jun 2006) $
	
--]]

local WEARME_NAME 			= "WearMe"
local WEARME_VERSION 		= 1.0
local WEARME_LAST_UPDATED	= "September 13, 2006"
local WEARME_AUTHOR 		= "AnduinLothar"
local WEARME_EMAIL			= "karlkfi@cosmosui.org"
local WEARME_WEBSITE		= "http://www.wowwiki.com/WearMe"

------------------------------------------------------------------------------
--[[ Embedded Sub-Library Load Algorithm ]]--
------------------------------------------------------------------------------

if (not WearMe) then
	WearMe = {};
end
local isBetterInstanceLoaded = ( WearMe.version and WearMe.version >= WEARME_VERSION );

if (not isBetterInstanceLoaded) then
	
	WearMe.version = WEARME_VERSION;
	
	------------------------------------------------------------------------------
	--[[ Globals ]]--
	------------------------------------------------------------------------------
	
	WearMe.DEBUG = false;
	
	--GetContainerItemLink item name strfind syntax
	WearMe.ITEM_NAME_FROM_LINK = "|h%[(.-)%]|h";
	WearMe.ITEM_UNIQUE_IDS_FROM_LINK = "|H(item:(%d-):(%d-):(%d-):(%d-))|h%[(.-)%]|h"
	
	WearMe.DuelInventorySlots = {
		[11] = 12;
		[12] = 11;
		[13] = 14;
		[14] = 13;
		[16] = 17;
		[17] = 16;
	};
	
	WearMe.DuelInventoryLocations = {
		["INVTYPE_FINGER"] = 11;
		["INVTYPE_TRINKET"] = 13;
		["INVTYPE_WEAPON"] = 16;
	};
	
	WearMe.BankBags = { -1, 5, 6, 7, 8, 9, 10 };
	
	WearMe.Containers = {
		[0] = {};
		[1] = {};
		[2] = {};
		[3] = {};
		[4] = {};
		[5] = {};
		[6] = {};
		[7] = {};
		[8] = {};
		[9] = {};
		[10] = {};
	};
	
	WearMe.Inventory = {};
	
	WearMe.CursorItem = {};
	
	WearMe.InventoryContainers = {};
	for i=1, 20 do
		WearMe.InventoryContainers[ContainerIDToInventoryID(i)] = i;
	end
	
	WearMe.MoveQueue = {};
	
	--[[ Inventory Slot Key:
	0 = ammo 
	1 = head 
	2 = neck 
	3 = shoulder 
	4 = shirt 
	5 = chest 
	6 = belt 
	7 = legs 
	8 = feet 
	9 = wrist 
	10 = gloves 
	11 = finger 1 
	12 = finger 2 
	13 = trinket 1 
	14 = trinket 2 
	15 = back 
	16 = main hand 
	17 = off hand 
	18 = ranged 
	19 = tabard
	]]--
	
	------------------------------------------------------------------------------
	--[[ Inventory Bag Item ID to Container ID Conversion ]]--
	------------------------------------------------------------------------------
	
	function WearMe.InventoryIDToContainerID(invSlot)
		-- Only works for bags (20-23, 64-69)
		if (invSlot) then
			return WearMe.InventoryContainers[invSlot];
		end
	end
		
	------------------------------------------------------------------------------
	--[[ Get Funcs ]]--
	------------------------------------------------------------------------------
	
	function WearMe.GetInventoryItemName(invSlot, forceCache)
		local item = WearMe.Inventory[invSlot];
		if (not forceCache and item and item.cached and item.name) then
			return item.name;
		else
			local linktext = GetInventoryItemLink("player", invSlot);
			if (linktext) then
				local _, _, itemName = strfind(linktext, WearMe.ITEM_NAME_FROM_LINK);
				WearMe.CacheInventoryItemName(invSlot, itemName);
				return itemName;
			end
		end
	end
	
	function WearMe.GetInventoryItemInfo(invSlot, forceCache)
		local item = WearMe.Inventory[invSlot];
		if (not forceCache and item and item.cached and item.itemID) then
			return item.itemString, item.itemID, item.enchant, item.suffix, item.bonus, item.name;
		else
			local linktext = GetInventoryItemLink("player", invSlot);
			if (linktext) then
				local _, _, itemString, itemID, permEnchant, suffix, extraItemInfo, itemName = strfind(linktext, WearMe.ITEM_UNIQUE_IDS_FROM_LINK);
				itemID = tonumber(itemID);
				permEnchant = tonumber(permEnchant);
				suffix = tonumber(suffix);
				extraItemInfo = tonumber(extraItemInfo);
				WearMe.CacheInventoryItem(invSlot, itemString, itemID, permEnchant, suffix, extraItemInfo, itemName);
				return itemString, itemID, permEnchant, suffix, extraItemInfo, itemName;
			end
		end
	end
	
	function WearMe.GetContainerItemName(bag, slot, forceCache)
		local item = WearMe.Containers[bag]
		if (item) then
			item = item[slot];
		end
		if (not forceCache and item and item.cached and item.name) then
			return item.name;
		else
			local linktext = GetContainerItemLink(bag, slot);
			if (linktext) then
				local _, _, itemName = strfind(linktext, WearMe.ITEM_NAME_FROM_LINK);
				WearMe.CacheContainerItemName(bag, slot, itemName);
				return itemName;
			end
		end
	end
	
	function WearMe.GetContainerItemInfo(bag, slot, forceCache)
		local item = WearMe.Containers[bag]
		if (item) then
			item = item[slot];
		end
		if (not forceCache and item and item.cached and item.itemID) then
			return item.itemString, item.itemID, item.enchant, item.suffix, item.bonus, item.name;
		else
			local linktext = GetContainerItemLink(bag, slot);
			if (linktext) then
				local _, _, itemString, itemID, permEnchant, suffix, extraItemInfo, itemName = strfind(linktext, WearMe.ITEM_UNIQUE_IDS_FROM_LINK);
				itemID = tonumber(itemID);
				permEnchant = tonumber(permEnchant);
				suffix = tonumber(suffix);
				extraItemInfo = tonumber(extraItemInfo);
				WearMe.CacheInventoryItem(bag, slot, itemString, itemID, permEnchant, suffix, extraItemInfo, itemName);
				return itemString, itemID, permEnchant, suffix, extraItemInfo, itemName;
			end
		end
	end
	
	
	------------------------------------------------------------------------------
	--[[ Find By Name ]]--
	------------------------------------------------------------------------------
	
	function WearMe.FindInventoryItemByName(targetItemName)
		-- Find the named item on the character's inventory (head slot, hand slot, etc -- not bags)
		for i = 0, 19 do
			if (targetItemName == WearMe.GetInventoryItemName(i)) then
				return i;
			end
		end
	end
	
	function WearMe.FindBagByName(targetBagName)
		-- For finding the bag item name, not the slot name
		-- Returns an Inventory ID, use InventoryIDToContainerID if you need the Container ID
		for i = 20, 23 do
			if (targetBagName == WearMe.GetInventoryItemName(i)) then
				return i;
			end
		end
	end
	
	function WearMe.FindBankBagByName(targetBagName)
		-- For finding the bag item name, not the slot name
		-- Only usable at the bank
		-- Returns an Inventory ID, use InventoryIDToContainerID if you need the Container ID
		for i = 64, 69 do
			if (targetBagName == WearMe.GetInventoryItemName(i)) then
				return i;
			end
		end
	end
	
	function WearMe.FindContainerItemByName(targetItemName)
		for bag = 0, 4 do
			for slot = 1, GetContainerNumSlots(bag) do
				if (targetItemName == WearMe.GetContainerItemName(bag, slot)) then
					return bag, slot;
				end
			end
		end
	end
	
	function WearMe.FindKeyByName(targetItemName)
		local bag = -2;
		for slot = 1, GetContainerNumSlots(bag) do
			if (targetItemName == WearMe.GetContainerItemName(bag, slot)) then
				return bag, slot;
			end
		end
	end
	
	function WearMe.FindBankItemByName(targetItemName)
		-- Only usable at the bank
		-- Searches through all the bank items using Container IDs
		for i, bag in ipairs(WearMe.BankBags) do
			for slot = 1, GetContainerNumSlots(bag) do
				if (targetItemName == WearMe.GetContainerItemName(bag, slot)) then
					return bag, slot;
				end
			end
		end
	end
	
	
	------------------------------------------------------------------------------
	--[[ Find By ID ]]--
	-- itemID can be an itemID or "itemString" (ex: 12345 or "item:12345:0:0:0")
	------------------------------------------------------------------------------
	
	-- Find the named item on the character's inventory (head slot, hand slot, etc -- not bags)
	function WearMe.FindInventoryItem(targetItemID)
		local itemString, itemID, permEnchant, suffix, extraItemInfo;
		for i = 0, 19 do
			itemString, itemID = WearMe.GetInventoryItemInfo(i);
			if (targetItemID == itemString or targetItemID == itemID) then
				return i;
			end
		end
	end
	
	function WearMe.FindBag(targetItemID)
		-- For finding the bag item name, not the slot name
		-- Returns an Inventory ID, use InventoryIDToContainerID if you need the Container ID
		local itemString, itemID, permEnchant, suffix, extraItemInfo;
		for i = 20, 23 do
			itemString, itemID = WearMe.GetInventoryItemInfo(i);
			if (targetItemID == itemString or targetItemID == itemID) then
				return i;
			end
		end
	end
	
	function WearMe.FindBankBag(targetItemID)
		-- For finding the bag item name, not the slot name
		-- Only usable at the bank
		-- Returns an Inventory ID, use InventoryIDToContainerID if you need the Container ID
		local itemString, itemID, permEnchant, suffix, extraItemInfo;
		for i = 64, 69 do
			itemString, itemID, permEnchant, suffix, extraItemInfo = WearMe.GetInventoryItemInfo(i);
			if (targetItemID == itemString or targetItemID == itemID) then
				return i;
			end
		end
	end
	
	function WearMe.FindContainerItem(targetItemID)
		local itemString, itemID, permEnchant, suffix, extraItemInfo;
		for bag = 0, 4 do
			for slot = 1, GetContainerNumSlots(bag) do
				itemString, itemID, permEnchant, suffix, extraItemInfo = WearMe.GetContainerItemInfo(bag, slot);
				if (targetItemID == itemString or targetItemID == itemID) then
					return bag, slot;
				end
			end
		end
	end
	
	function WearMe.FindKey(targetItemID)
		local itemString, itemID, permEnchant, suffix, extraItemInfo;
		local bag = -2;
		for slot = 1, GetContainerNumSlots(bag) do
			itemString, itemID, permEnchant, suffix, extraItemInfo = WearMe.GetContainerItemInfo(bag, slot);
			if (targetItemID == itemString or targetItemID == itemID) then
				return bag, slot;
			end
		end
	end
	
	function WearMe.FindBankItem(targetItemID)
		-- Only usable at the bank
		local itemString, itemID, permEnchant, suffix, extraItemInfo;
		for i, bag in ipairs(WearMe.BankBags) do
			for slot = 1, GetContainerNumSlots(bag) do
				itemString, itemID, permEnchant, suffix, extraItemInfo = WearMe.GetContainerItemInfo(bag, slot);
				if (targetItemID == itemString or targetItemID == itemID) then
					return bag, slot;
				end
			end
		end
	end
	
	
	------------------------------------------------------------------------------
	--[[ Find By ID or Name ]]--
	------------------------------------------------------------------------------
	
	-- Find the item on the character's inventory or bags
	function WearMe.PlayerHasItem(itemID, itemName)
		local invSlot, bag;
		if (itemID) then
			-- Check inventory
			invSlot = WearMe.FindInventoryItem(itemID);
			if (invSlot) then
				-- Item equipped
				return true;
			else
				-- Check bags
				bag = WearMe.FindContainerItem(itemID);
				if (bag) then
					-- Item in bag
					return true;
				end
			end
		end
		if (itemName) then
			-- Check inventory
			invSlot = WearMe.FindInventoryItemByName(itemName);
			if (invSlot) then
				-- Item equipped
				return true;
			else
				-- Check bag slots
				bag = WearMe.FindContainerItemByName(itemName);
				if (bag) then
					-- Item in bag
					return true;
				end
			end
		end
		return false;
	end
	
	
	function WearMe.PlayerHasInventoryItem(itemID, itemName)
		local invSlot;
		if (itemID) then
			-- Check inventory
			invSlot = WearMe.FindInventoryItem(itemID);
			if (invSlot) then
				-- Item equipped
				return true;
			end
		end
		if (itemName) then
			-- Check inventory
			invSlot = WearMe.FindInventoryItemByName(itemName);
			if (invSlot) then
				-- Item equipped
				return true;
			end
		end
		return false;
	end
	
	
	function WearMe.PlayerHasContainerItem(itemID, itemName)
		local bag;
		if (itemID) then
			-- Check inventory
			bag = WearMe.FindContainerItem(itemID);
			if (bag) then
				-- Item equipped
				return true;
			end
		end
		if (itemName) then
			-- Check inventory
			bag = WearMe.FindContainerItemByName(itemName);
			if (bag) then
				-- Item equipped
				return true;
			end
		end
		return false;
	end
	
	---------------------------------------------------------------------------------
	-- Put the item on the cursor into a free bag slot
	---------------------------------------------------------------------------------
	
	function WearMe.BagCursorItem(backwards)
		if (CursorHasItem()) then
			if (backwards) then
				for bag = 4, 0, -1 do
					for slot = GetContainerNumSlots(bag), 1, -1 do
						if (not WearMe.ContainerSlotInUse(bag, slot)) then
							PickupContainerItem(bag, slot);
							return bag, slot;
						end
					end
				end
			else
				for bag = 0, 4 do
					for slot = 1, GetContainerNumSlots(bag) do
						if (not WearMe.ContainerSlotInUse(bag, slot)) then
							--Sea.io.printComma(bag, slot)
							PickupContainerItem(bag, slot);
							return bag, slot;
						end
					end
				end
			end
		end
	end
	
	
	---------------------------------------------------------------------------------
	-- Move an item to the front most availible bag (or furthest back)
	---------------------------------------------------------------------------------
	
	function WearMe.ReBagContainerItem(bag, slot, backwards)
		if (backwards) then
			for targetBag = 4, bag+1, -1 do
				for targetSlot = GetContainerNumSlots(targetBag), 1, -1 do
					if (not WearMe.ContainerSlotInUse(targetBag, targetSlot)) then
						PickupContainerItem(bag, slot);
						if (not CursorHasItem()) then
							return;
						end
						PickupContainerItem(targetBag, targetSlot);
						return bag, slot;
					end
				end
			end
		else
			for targetBag = 0, bag-1 do
				for targetSlot = 1, GetContainerNumSlots(targetBag) do
					if (not WearMe.ContainerSlotInUse(targetBag, targetSlot)) then
						PickupContainerItem(bag, slot);
						if (not CursorHasItem()) then
							return;
						end
						PickupContainerItem(targetBag, targetSlot);
						return bag, slot;
					end
				end
			end
		end
	end
	
	
	function WearMe.ReBagContainerItemByName(itemName, backwards)
		local bag, slot = WearMe.FindContainerItemByName(itemName);
		if (bag) then
			return WearMe.ReBagContainerItem(bag, slot, backwards);
		end
	end
	
	function WearMe.ReBagContainerItemByID(itemID, backwards)
		local bag, slot = WearMe.FindContainerItem(itemID);
		if (bag) then
			return WearMe.ReBagContainerItem(bag, slot, backwards);
		end
	end
	
	
	---------------------------------------------------------------------------------
	-- Equiping
	---------------------------------------------------------------------------------
	-- Event error if there's enough bag slots when equipping a 2h from 2 1h
	-- secondarySlot is ignored if it's not a 2 slotted item type
	-- Event error if you try to equip a main hand weapon to the secondarySlot
	
	function WearMe.EquipByName(targetItemName, secondarySlot)
		-- if we're already holding something, bail
		if (CursorHasItem()) then
			return false;
		end
		
		-- Check if it's already equipped and possibly in an alternate spot
		local _, itemName;
		for invSlot, altSlot in pairs(WearMe.DuelInventorySlots) do
			itemName = WearMe.GetInventoryItemName(invSlot);
			if (itemName and itemName == targetItemName) then
				if (secondarySlot and altSlot < invSlot) or (not secondarySlot and altSlot > invSlot) then
					return true;
				else
					PickupInventoryItem(invSlot);
					EquipCursorItem(altSlot);
					return true;
				end
			end
		end
		
		-- Check non-duplicate inventory slots (duplicate slots will never return possitive)
		local invSlot = WearMe.FindInventoryItemByName(targetItemName);
		if (invSlot) then
			-- Item already equipped
			return true;
		else
			-- Check bag slots
			local bag, bagSlot = WearMe.FindContainerItemByName(targetItemName);
			if (not bag) then
				-- Item Name not found in inventory or bag
				return false;
			end
			PickupContainerItem(bag, bagSlot);
		end
		
		if (secondarySlot) then
			local _, _, _, _, _, _, _, itemEquipLoc = GetItemInfo(WearMe.CursorItem.itemString);
			if (itemEquipLoc == "INVTYPE_FINGER") then
				EquipCursorItem(12);
				return true;
			elseif (itemEquipLoc == "INVTYPE_TRINKET") then
				EquipCursorItem(14);
				return true;
			elseif (itemEquipLoc == "INVTYPE_WEAPON" or itemEquipLoc == "INVTYPE_SHIELD" or itemEquipLoc == "INVTYPE_WEAPONOFFHAND" or itemEquipLoc == "INVTYPE_HOLDABLE") then
				EquipCursorItem(17);
				return true;
			end
		end
		
		AutoEquipCursorItem();
		return true;
	end
	
	
	-- itemID can be an itemID or "itemString" (ex: 12345 or "item:12345:0:0:0")
	function WearMe.Equip(targetItemID, secondarySlot)
		-- if we're already holding something, bail
		if (CursorHasItem()) then
			return false;
		end
		
		local _, _, _, _, _, _, _, itemEquipLoc = GetItemInfo(targetItemID);
		
		local invSlot = WearMe.DuelInventoryLocations[itemEquipLoc];
		if (invSlot) then
			local altSlot = invSlot + 1;
			-- Check if it's already equipped and possibly in an alternate spot
			local itemString, itemID = WearMe.GetInventoryItemInfo(invSlot);
			if (itemString) and (itemString == targetItemID or itemID == targetItemID) then
				if (not secondarySlot) then
					return true;
				else
					PickupInventoryItem(invSlot);
					EquipCursorItem(altSlot);
					return true;
				end
			end
			itemString, itemID = WearMe.GetInventoryItemInfo(altSlot);
			if (itemString) and (itemString == targetItemID or itemID == targetItemID) then
				if (secondarySlot) then
					return true;
				else
					PickupInventoryItem(altSlot);
					EquipCursorItem(invSlot);
					return true;
				end
			end
		end
		
		-- Check non-duplicate inventory slots (duplicate slots will never return possitive)
		invSlot = WearMe.FindInventoryItem(targetItemID);
		if (invSlot) then
			-- Item already equipped
			return true;
		else
			-- Check bag slots
			local bag, bagSlot = WearMe.FindContainerItem(targetItemID);
			if (not bag) then
				-- Item Name not found in inventory or bag
				return false;
			end
			local itemString = WearMe.GetContainerItemInfo(bag, bagSlot);
			PickupContainerItem(bag, bagSlot);
		end
		
		if (secondarySlot) then
			if (itemEquipLoc == "INVTYPE_FINGER") then
				EquipCursorItem(12);
				return true;
			elseif (itemEquipLoc == "INVTYPE_TRINKET") then
				EquipCursorItem(14);
				return true;
			elseif (itemEquipLoc == "INVTYPE_WEAPON" or itemEquipLoc == "INVTYPE_SHIELD" or itemEquipLoc == "INVTYPE_WEAPONOFFHAND" or itemEquipLoc == "INVTYPE_HOLDABLE") then
				EquipCursorItem(17);
				return true;
			end
		end
		
		AutoEquipCursorItem();
		return true;
	end
	
	
	---------------------------------------------------------------------------------
	-- Unequiping
	---------------------------------------------------------------------------------
	
	function WearMe.Unequip(invSlot, backwards)
		local texture = GetInventoryItemTexture("player", invSlot);
		if (texture) then
			PickupInventoryItem(invSlot);
			if (WearMe.BagCursorItem(backwards)) then
				ClearCursor();
				return false;
			end
		end
		return true;
	end
	
	
	---------------------------------------------------------------------------------
	-- Check if there's an item in an inventory or bag slot
	---------------------------------------------------------------------------------
	
	function WearMe.InventorySlotInUse(invSlot, forceCache)
		local item = WearMe.Inventory[invSlot];
		if (not forceCache and item and item.cached) then
			if (item.name) then
				return true;
			end
		else
			if (WearMe.GetInventoryItemInfo(invSlot, true)) then
				return true;
			end
		end
	end
	
	function WearMe.ContainerSlotInUse(bag, slot, forceCache)
		if (not WearMe.Containers[bag]) then
			WearMe.Containers[bag] = {};
		end
		local item = WearMe.Containers[bag][slot];
		if (not forceCache and item and item.cached) then
			if (item.name) then
				return true;
			end
		else
			if (WearMe.GetContainerItemInfo(bag, slot, true)) then
				return true;
			end
		end
	end
	
	function WearMe.GetNumEmptyContainerSlots(forceCache)
		local bag, item;
		local numSlots = 0;
		for bagNum=0, 4 do
			bag = WearMe.Containers[bagNum];
			if (not bag) then
				WearMe.Containers[bagNum] = {};
				bag = WearMe.Containers[bagNum];
			end
			for bagSlot = 1, GetContainerNumSlots(bagNum) do
				item = bag[bagSlot];
				if (not forceCache and item and item.cached) then
					if (not item.name) then
						numSlots = numSlots + 1;
					end
				else
					if (not WearMe.GetContainerItemInfo(bagNum, bagSlot, true)) then
						numSlots = numSlots + 1;
					end
				end
			end
		end
		return numSlots;
	end
	
	---------------------------------------------------------------------------------
	--[[ Optimized Outfit Swapping ]]--
	-- Only swaps what needs to be swapped.
	-- Swaps, even if it's in another slot.
	-- Moves unequipped offhands, trinkets and rings to the back most bag (if swapping in a two hand or moving an equipped item to its alternate slot).
	-- Checks for id first. If that id is not found it will check for the name. This lets you prefer items with enchants or suffixes.
	-- Allows you to specify in any order.
	-- Multiple registers for the same slot will use the most recent.
	-- Outfit is preserved until the next init, so you can WearOutfit later without reregistering.
	-- If there is not enough space in your bags or the item is not found, RegisterToEquip returns false.
	--[[
	EX:
	WearMe.InitOutfit();
	WearMe.RegisterToEquip(16, "item:18822:2646:0:0", "Obsidian Edged Blade");	-- Equip OEB w/ +25 agil (by itemString) or OEB (by name)
	WearMe.RegisterToEquip(11, 19325);		-- Equip Ring 1 by id
	WearMe.RegisterToEquip(12, 19898);		-- Equip Ring 2 by id
	WearMe.RegisterToEquip(19);				-- Unequip Tabard
	WearMe.WearOutfit();
	]]--
	---------------------------------------------------------------------------------
	
	WearMe.OutfitItems = {};
	for i=0, 19 do
		WearMe.OutfitItems[i] = {};
	end
	WearMe.OutfitItems.reqBagSlots = 0;
	
	
	function WearMe.InitOutfit()
		local item;
		for i=0, 19 do
			item = WearMe.OutfitItems[i];
			if (not item) then
				WearMe.OutfitItems[i] = {};
				item = WearMe.OutfitItems[i];
			end
			item.id = nil;
			item.name = nil;
			item.twohand = nil;
			item.specified = nil;
		end
		WearMe.OutfitItems.reqBagSlots = 0;
	end
	
	
	function WearMe.RegisterToEquip(targetInvSlot, targetItemID, targetItemName)
		
		local item = WearMe.OutfitItems[targetInvSlot];
		local invSlot, bag, bagSlot
		
		if (not targetItemID and not targetItemName) then
			-- Make sure there's enough room to unequip
			if (WearMe.InventorySlotInUse(targetInvSlot)) then
				if (WearMe.GetNumEmptyContainerSlots() <= WearMe.OutfitItems.reqBagSlots) then
					return false;
				end
			end
			WearMe.OutfitItems.reqBagSlots = WearMe.OutfitItems.reqBagSlots + 1;
			item.specified = true;
			return true;
		end
		
		if (targetItemID) then
			invSlot = WearMe.FindInventoryItem(targetItemID);
			if (invSlot) then
				if (invSlot == targetInvSlot) then
					return true;
				end
				item.id = targetItemID;
				item.specified = true;
			else
				bag, bagSlot = WearMe.FindContainerItem(targetItemID);
				if (bag) then
					item.id = targetItemID;
					item.specified = true;
				end
			end
		end
		
		-- If both id and name were suplied, but the item was not found by that id, try by name
		if (targetItemName and not item.specified) then
			invSlot = WearMe.FindInventoryItemByName(targetItemName)
			if (invSlot) then
				if (invSlot == targetInvSlot) then
					return true;
				end
				item.name = targetItemName;
				item.specified = true;
			else
				bag, bagSlot = WearMe.FindContainerItemByName(targetItemName);
				if (bag) then
					item.name = targetItemName;
					item.specified = true;
				end
			end
		end
		
		if (not item.specified) then
			return false;
		end
		
		if (item.specified and targetInvSlot == 16) then	-- Main Hand
			if (invSlot) then
				local itemString = WearMe.GetInventoryItemInfo(invSlot);
				local _, _, _, _, _, _, _, equipLoc = GetItemInfo(itemString);
				if (equipLoc == "INVTYPE_2HWEAPON") then
					item.twohand = true;
				end
			elseif (bag) then
				local itemString = WearMe.GetContainerItemInfo(bag, bagSlot);
				local _, _, _, _, _, _, _, equipLoc = GetItemInfo(itemString);
				if (equipLoc == "INVTYPE_2HWEAPON") then
					item.twohand = true;
				end
			end
		end
		
		return true;
	end
	
	
	function WearMe.WearOutfit()
		
		local outfit = WearMe.OutfitItems;
		
		-- Rings: 11, 12
		-- Trinkets: 13, 14
		-- Weapons: 16, 17
		local _, itemString, itemID, itemName, altItemString, altItemID, altItemName;
		for invSlot, altSlot in pairs(WearMe.DuelInventorySlots) do
			
			-- Ignore secondary slots (handle them together on the primary slot)
			if (invSlot < altSlot) then
			
				if (invSlot == 16 and outfit[invSlot].twohand) then
					WearMe.WearItem(outfit[invSlot], invSlot);
					altItemString, altItemID, _, _, _, altItemName = WearMe.GetInventoryItemInfo(altSlot);
					if (altItemString) then
						WearMe.QueueItemForMoving(altItemString);
					end
						
				elseif (outfit[invSlot].specified and outfit[altSlot].specified and WearMe.InventorySlotInUse(invSlot) and WearMe.InventorySlotInUse(altSlot)) then
					
					itemString, itemID, _, _, _, itemName = WearMe.GetInventoryItemInfo(invSlot);
					altItemString, altItemID, _, _, _, altItemName = WearMe.GetInventoryItemInfo(altSlot);
					
					if (not itemString or not altItemString) then
						WearMe.WearItem(outfit[invSlot], invSlot);
						WearMe.WearItem(outfit[altSlot], altSlot);
						
					elseif (outfit[altSlot].name == itemName or outfit[altSlot].id == itemString or outfit[altSlot].id == itemID)
							and (outfit[invSlot].name == altItemName or outfit[invSlot].id == altItemString or outfit[invSlot].id == altItemID) then
						-- Both equipped, just in the wrong slots
						PickupInventoryItem(invSlot);
						EquipCursorItem(altSlot);
						
					elseif (outfit[altSlot].name == itemName or outfit[altSlot].id == itemString or outfit[altSlot].id == itemID) then
						-- Primary slot item needs to go into the secondary slot
						-- Put secondary slot item in bag
						WearMe.Unequip(altSlot, true);
						WearMe.QueueItemForMoving(altItemString);
						-- Equip primary slot item to secondary slot
						PickupInventoryItem(invSlot);
						EquipCursorItem(altSlot);
						-- Equip the desired item to the primary slot
						if (outfit[invSlot].name or outfit[invSlot].id) then
							WearMe.WearItem(outfit[invSlot], invSlot);
						end
						
					else
						WearMe.WearItem(outfit[invSlot], invSlot);
						WearMe.WearItem(outfit[altSlot], altSlot);
						
					end
					
				else
					-- Either one slot is empty or only one slot is specified for the outfit
					if (outfit[invSlot].specified) then
						WearMe.WearItem(outfit[invSlot], invSlot);
					end
					if (outfit[altSlot].specified) then
						WearMe.WearItem(outfit[altSlot], altSlot);
					end
					
				end
				
			end
			
		end
		
		-- All Slots except the duplicates (previously handled as exceptions)
		for i=0, 19 do
			if (not WearMe.DuelInventorySlots[i]) then
				WearMe.WearItem(outfit[i], i);
			end
		end
			
	end
	
	
	function WearMe.WearItem(item, invSlot)
		local altSlot = WearMe.DuelInventorySlots[invSlot];
		local secondarySlot = altSlot and altSlot < invSlot;
		if (item.name) then
			WearMe.EquipByName(item.name, secondarySlot);
		elseif (item.id) then
			WearMe.Equip(item.id, secondarySlot);
		elseif (item.specified) then
			 WearMe.Unequip(invSlot, true);
		end
	end
	
	
	------------------------------------------------------------------------------
	--[[ Move to back of bags queue ]]--
	------------------------------------------------------------------------------
	
	function WearMe.QueueItemForMoving(itemID)
		if (not WearMe.MoveQueue) then
			WearMe.MoveQueue = {};
		end
		tinsert(WearMe.MoveQueue, itemID);
	end
	
	
	function WearMe.CallMoveQueue()
		if (WearMe.MoveQueue and getn(WearMe.MoveQueue) > 0) then
			for i=getn(WearMe.MoveQueue), 1, -1 do
				if (WearMe.ReBagContainerItemByID(WearMe.MoveQueue[i], true)) then
					table.remove(WearMe.MoveQueue, i);
				end
			end
		end
	end
	
	
	------------------------------------------------------------------------------
	--[[ Inventory Cache ]]--
	------------------------------------------------------------------------------
	
	function WearMe.ClearInventoryCache()
		local item;
		for invSlot = 0, 19 do
			item = WearMe.Inventory[invSlot];
			if (not item) then
				WearMe.Inventory[invSlot] = {};
			else
				item.itemString = nil;
				item.itemID = nil;
				item.enchant = nil;
				item.suffix = nil;
				item.bonus = nil;
				item.name = nil;
				item.cached = nil;
			end
		end
	end
	
	function WearMe.ClearContainerCache(bag)
		local to, from;
		if (bag) then
			to = bag;
			from = bag;
		else
			to = 10;
			from = 0;
		end
		local item;
		for bag = from, to do
			for slot = 1, GetContainerNumSlots(bag) do
				item = WearMe.Containers[bag][slot];
				if (not item) then
					WearMe.Containers[bag][slot] = {};
				else
					item.itemString = nil;
					item.itemID = nil;
					item.enchant = nil;
					item.suffix = nil;
					item.bonus = nil;
					item.name = nil;
					item.cached = nil;
				end
			end
		end
	end
	
	function WearMe.CacheInventoryItemName(invSlot, name)
		local item = WearMe.Inventory[invSlot];
		if (not item) then
			WearMe.Inventory[invSlot] = {};
			item = WearMe.Inventory[invSlot];
		end
		if (type(name) ~= "number") then
			if (name ~= item.name) then
				item.itemString = nil;
				item.itemID = nil;
				item.enchant = nil;
				item.suffix = nil;
				item.bonus = nil;
			end
			item.name = name;
			item.cached = true;
		end
	end
	
	function WearMe.CacheInventoryItem(invSlot, itemString, itemID, permEnchant, suffix, extraItemInfo, itemName)
		local item = WearMe.Inventory[invSlot];
		if (not item) then
			WearMe.Inventory[invSlot] = {};
			item = WearMe.Inventory[invSlot];
		end
		if (type(itemString) ~= "number") then
			item.itemString = itemString;
			item.itemID = itemID;
			item.enchant = permEnchant;
			item.suffix = suffix;
			item.bonus = extraItemInfo;
			item.name = itemName;
			item.cached = true;
		else
			--Sea.io.printComma("Wrong itemString: ", invSlot, itemString, itemID, permEnchant, suffix, extraItemInfo, itemName)
		end
	end
	
	function WearMe.CacheContainerItemName(bag, slot, name)
		if (not WearMe.Containers[bag]) then
			WearMe.Containers[bag] = {};
		end
		local item = WearMe.Containers[bag][slot];
		if (not item) then
			WearMe.Containers[bag][slot] = {};
			item = WearMe.Containers[bag][slot];
		end
		if (type(name) ~= "number") then
			if (name ~= item.name) then
				item.itemString = nil;
				item.itemID = nil;
				item.enchant = nil;
				item.suffix = nil;
				item.bonus = nil;
			end
			item.name = name;
			item.cached = true;
		end
	end
	
	function WearMe.CacheContainerItem(bag, slot, itemString, itemID, permEnchant, suffix, extraItemInfo, itemName)
		if (not WearMe.Containers[bag]) then
			WearMe.Containers[bag] = {};
		end
		local item = WearMe.Containers[bag][slot];
		if (not item) then
			WearMe.Containers[bag][slot] = {};
			item = WearMe.Containers[bag][slot];
		end
		if (type(itemString) ~= "number") then
			--Sea.io.printComma("CacheContainerItem: setting ", bag, slot, itemName);
			item.itemString = itemString;
			item.itemID = itemID;
			item.enchant = permEnchant;
			item.suffix = suffix;
			item.bonus = extraItemInfo;
			item.name = itemName;
			item.cached = true;
		end
	end
	
	function WearMe.CacheCursorItem(itemString, itemID, permEnchant, suffix, extraItemInfo, itemName)
		local item = WearMe.CursorItem;
		if (not item) then
			WearMe.CursorItem = {};
			item = WearMe.CursorItem;
		end
		item = WearMe.CursorItem.nextItem;
		if (not item) then
			WearMe.CursorItem.nextItem = {};
			item = WearMe.CursorItem.nextItem;
		end
		--Sea.io.print("CacheCursorItem: setting nextItem: ", itemName);
		item.itemString = itemString;
		item.itemID = itemID;
		item.enchant = permEnchant;
		item.suffix = suffix;
		item.bonus = extraItemInfo;
		item.name = itemName;
		
		item.bag = nil;
		item.bagSlot = nil;
		item.invSlot = nil;
	end
	
	function WearMe.UpdateCursorCache()
		local item = WearMe.CursorItem;
		if (not item) then
			WearMe.CursorItem = {};
			item = WearMe.CursorItem;
		end
		
		if (item.nextItem and item.nextItem.name) then
			--Sea.io.print("UpdateCursorCache: copy from nextItem");
			item.itemString = item.nextItem.itemString;
			item.itemID = item.nextItem.itemID;
			item.enchant = item.nextItem.enchant;
			item.suffix = item.nextItem.suffix;
			item.bonus = item.nextItem.bonus;
			item.name = item.nextItem.name;
			
			item.bag = item.nextItem.bag;
			item.bagSlot = item.nextItem.bagSlot;
			item.invSlot = item.nextItem.invSlot;
			
			WearMe.ClearNextItemCursorCache();
		
		elseif (not CursorHasItem()) then
			--Sea.io.print("UpdateCursorCache: clearing cursor");
			item.itemString = nil;
			item.itemID = nil;
			item.enchant = nil;
			item.suffix = nil;
			item.bonus = nil;
			item.name = nil;
			
			item.bag = nil;
			item.bagSlot = nil;
			item.invSlot = nil;
			
			WearMe.ClearNextItemCursorCache();
			
		end
	end
	
	function WearMe.ClearNextItemCursorCache()
		--Sea.io.print("ClearNextItemCursorCache");
		if (WearMe.CursorItem and WearMe.CursorItem.nextItem) then
			local item = WearMe.CursorItem.nextItem;
			item.itemString = nil;
			item.itemID = nil;
			item.enchant = nil;
			item.suffix = nil;
			item.bonus = nil;
			item.name = nil;
			
			item.bag = nil;
			item.bagSlot = nil;
			item.invSlot = nil;
		end
	end
	
	
	------------------------------------------------------------------------------
	--[[ Cache Hooks ]]--
	------------------------------------------------------------------------------
	
	function WearMe.PickupContainerItem_Hook(bag, slot)
		if (CursorHasItem()) then
			local cursorItem = WearMe.CursorItem;
			if (cursorItem and cursorItem.itemID) then
				local itemString, itemID, permEnchant, suffix, extraItemInfo, itemName = WearMe.GetContainerItemInfo(bag, slot);
				--Sea.io.print("PickupContainerItem: Current Item: ", itemName, " Cursor Item: ", cursorItem.name)
				WearMe.CacheContainerItem(bag, slot, cursorItem.itemString, cursorItem.itemID, cursorItem.enchant, cursorItem.suffix, cursorItem.bonus, cursorItem.name);
				
				if (cursorItem.invSlot) then
					WearMe.CacheInventoryItem(cursorItem.invSlot, itemString, itemID, permEnchant, suffix, extraItemInfo, itemName);
				elseif (cursorItem.bag and cursorItem.bagSlot) then
					WearMe.CacheContainerItem(cursorItem.bag, cursorItem.bagSlot, itemString, itemID, permEnchant, suffix, extraItemInfo, itemName);
				end
			end
		else
			local itemString, itemID, permEnchant, suffix, extraItemInfo, itemName = WearMe.GetContainerItemInfo(bag, slot);
			WearMe.CacheCursorItem(itemString, itemID, permEnchant, suffix, extraItemInfo, itemName);
			WearMe.CursorItem.nextItem.bag = bag;
			WearMe.CursorItem.nextItem.bagSlot = bagSlot;
		end
	end
	
	Sea.util.unhook("PickupContainerItem", "WearMe.PickupContainerItem_Hook", "before");
	Sea.util.hook("PickupContainerItem", "WearMe.PickupContainerItem_Hook", "before");
	
	
	function WearMe.PickupInventoryItem_Hook(invSlot)
		if (CursorHasItem()) then
			local cursorItem = WearMe.CursorItem;
			if (cursorItem and cursorItem.itemID) then
				local itemString, itemID, permEnchant, suffix, extraItemInfo, itemName = WearMe.GetInventoryItemInfo(invSlot);
				WearMe.CacheInventoryItem(invSlot, cursorItem.itemString, cursorItem.itemID, cursorItem.enchant, cursorItem.suffix, cursorItem.bonus, cursorItem.name);
				
				if (cursorItem.invSlot) then
					WearMe.CacheInventoryItem(cursorItem.invSlot, itemString, itemID, permEnchant, suffix, extraItemInfo, itemName);
				elseif (cursorItem.bag and cursorItem.bagSlot) then
					WearMe.CacheContainerItem(cursorItem.bag, cursorItem.bagSlot, itemString, itemID, permEnchant, suffix, extraItemInfo, itemName);
				end
			end
		else
			local itemString, itemID, permEnchant, suffix, extraItemInfo, itemName = WearMe.GetInventoryItemInfo(invSlot);
			WearMe.CacheCursorItem(itemString, itemID, permEnchant, suffix, extraItemInfo, itemName);
			WearMe.CursorItem.nextItem.invSlot = invSlot;
		end
	end
	
	Sea.util.unhook("PickupInventoryItem", "WearMe.PickupInventoryItem_Hook", "before");
	Sea.util.hook("PickupInventoryItem", "WearMe.PickupInventoryItem_Hook", "before");
	
	function WearMe.EquipCursorItem_Hook(invSlot)
		if (CursorHasItem()) then
			local cursorItem = WearMe.CursorItem;
			if (cursorItem and cursorItem.itemID) then
				local itemString, itemID, permEnchant, suffix, extraItemInfo, itemName = WearMe.GetInventoryItemInfo(invSlot);
				WearMe.CacheInventoryItem(invSlot, cursorItem.itemString, cursorItem.itemID, cursorItem.enchant, cursorItem.suffix, cursorItem.bonus, cursorItem.name);
				
				if (cursorItem.invSlot) then
					WearMe.CacheInventoryItem(cursorItem.invSlot, itemString, itemID, permEnchant, suffix, extraItemInfo, itemName);
				elseif (cursorItem.bag and cursorItem.bagSlot) then
					WearMe.CacheContainerItem(cursorItem.bag, cursorItem.bagSlot, itemString, itemID, permEnchant, suffix, extraItemInfo, itemName);
				end
			end
		end
	end
	
	Sea.util.unhook("EquipCursorItem", "WearMe.EquipCursorItem_Hook", "before");
	Sea.util.hook("EquipCursorItem", "WearMe.EquipCursorItem_Hook", "before");
	
	
	------------------------------------------------------------------------------
	--[[ Event Handler ]]--
	------------------------------------------------------------------------------
	
	function WearMe.OnEvent()
		if (event == "BAG_UPDATE") then
			-- arg1 = Container ID
			WearMe.ClearContainerCache(arg1);
			--Sea.io.print("BAG_UPDATE: ", arg1);
			
			
		elseif (event == "UNIT_INVENTORY_CHANGED") then
			if (arg1 == "player") then
				--Sea.io.print("UNIT_INVENTORY_CHANGED: ", arg1);
				WearMe.ClearInventoryCache();
				WearMe.CallMoveQueue();
			end
		
		elseif (event == "CURSOR_UPDATE") then
			--Sea.io.print("CURSOR_UPDATE: ", CursorHasItem());
			WearMe.UpdateCursorCache();
		
		end
	end
	
	------------------------------------------------------------------------------
	--[[ Event Driver ]]--
	------------------------------------------------------------------------------

	if (not WearMeFrame) then
		CreateFrame("Frame", "WearMeFrame");
	end
	WearMeFrame:Hide();
	--Frame Scripts
	WearMeFrame:SetScript("OnEvent", WearMe.OnEvent);
	WearMeFrame:RegisterEvent("BAG_UPDATE");
	WearMeFrame:RegisterEvent("UNIT_INVENTORY_CHANGED");
	WearMeFrame:RegisterEvent("CURSOR_UPDATE");
	
end
